from __future__ import annotations
import math
from typing import Any, Callable, Iterable

import torch
from torch import Tensor
from torch.optim.optimizer import Optimizer

from ranzen.decorators import implements

__all__ = ["RAdam"]


class RAdam(Optimizer):
    """Implements the Rectified Adam (RAdam) algorithm."""

    def __init__(
        self,
        params: Iterable[Tensor | dict[str, Any]],
        lr: float = 1e-3,
        betas: tuple[float, float] = (0.9, 0.999),
        eps: float = 1e-6,
        weight_decay: float = 0.0,
    ) -> None:
        """
        :param params: iterable of parameters to optimize or dicts defining parameter groups.
        :param lr: learning rate.
        :param betas: coefficients used for computing running averages of gradient and its square.
        :param eps: term added to the denominator to improve numerical stability.
        :param weight_decay: weight decay coefficient.

        :raises ValueError: if any one of ``lr``, ``betas``, ``eps``, or ``weight_decay`` is not in
            its permitted range.
        """
        if lr <= 0.0:
            raise ValueError(f"Invalid learning rate: {lr}")
        if eps < 0.0:
            raise ValueError(f"Invalid epsilon value: {eps}")
        if not 0.0 <= betas[0] < 1.0:
            raise ValueError(f"Invalid beta parameter at index 0: {betas[0]}")
        if not 0.0 <= betas[1] < 1.0:
            raise ValueError(f"Invalid beta parameter at index 1: {betas[1]}")
        if weight_decay < 0:
            raise ValueError(f"Invalid weight_decay value: {weight_decay}")

        buffer = [[None, None, None]] * 10
        if isinstance(params, (list, tuple)) and (len(params) > 0):
            for param in params:
                if isinstance(param, dict):
                    if "betas" in param and (
                        (param["betas"][0] != betas[0]) or (param["betas"][1] != betas[1])
                    ):
                        param["buffer"] = buffer

        defaults = dict(lr=lr, betas=betas, eps=eps, weight_decay=weight_decay, buffer=buffer)
        super().__init__(params=params, defaults=defaults)  # type: ignore

    @implements(Optimizer)
    def step(self, closure: Callable[[], Tensor] | None = None) -> Tensor | None:
        r"""Performs a single optimization step.

        :param closure: A closure that reevaluates the model and returns the loss.
        :returns: loss returned by the closure if ``closure`` is not ``None`` else ``None``.

        :raises RuntimeError: if gradients are sparse.
        """
        loss = None
        if closure is not None:
            loss = closure()

        for group in self.param_groups:

            for p in group["params"]:
                if p.grad is None:
                    continue
                grad = p.grad.data.float()
                if grad.is_sparse:
                    raise RuntimeError("RAdam does not support sparse gradients")

                p_data_fp32 = p.data.float()

                state = self.state[p]

                if len(state) == 0:
                    state["step"] = 0
                    state["exp_avg"] = torch.zeros_like(p_data_fp32)
                    state["exp_avg_sq"] = torch.zeros_like(p_data_fp32)
                else:
                    state["exp_avg"] = state["exp_avg"].type_as(p_data_fp32)
                    state["exp_avg_sq"] = state["exp_avg_sq"].type_as(p_data_fp32)

                exp_avg, exp_avg_sq = state["exp_avg"], state["exp_avg_sq"]
                beta1, beta2 = group["betas"]

                exp_avg_sq.mul_(beta2).addcmul_(grad, grad, value=1 - beta2)
                exp_avg.mul_(beta1).add_(grad, alpha=1 - beta1)

                state["step"] += 1
                buffered = group["buffer"][int(state["step"] % 10)]
                if state["step"] == buffered[0]:
                    n_sma, step_size = buffered[1], buffered[2]
                else:
                    buffered[0] = state["step"]
                    beta2_t = beta2 ** state["step"]
                    n_sma_max = 2 / (1 - beta2) - 1
                    n_sma = n_sma_max - 2 * state["step"] * beta2_t / (1 - beta2_t)
                    buffered[1] = n_sma

                    # more conservative since it's an approximated value
                    if n_sma >= 5:
                        step_size = math.sqrt(
                            (1 - beta2_t)
                            * (n_sma - 4)
                            / (n_sma_max - 4)
                            * (n_sma - 2)
                            / n_sma
                            * n_sma_max
                            / (n_sma_max - 2)
                        ) / (1 - beta1 ** state["step"])
                    else:
                        step_size = -1
                    buffered[2] = step_size

                # more conservative since it's an approximated value
                if n_sma >= 5:
                    if group["weight_decay"] != 0:
                        p_data_fp32.add_(-group["weight_decay"] * group["lr"], p_data_fp32)
                    denom = exp_avg_sq.sqrt().add_(group["eps"])
                    p_data_fp32.addcdiv_(-step_size * group["lr"], exp_avg, denom)
                    p.data.copy_(p_data_fp32)
                elif step_size > 0:
                    if group["weight_decay"] != 0:
                        p_data_fp32.add_(-group["weight_decay"] * group["lr"], p_data_fp32)
                    p_data_fp32.add_(-step_size * group["lr"], exp_avg)
                    p.data.copy_(p_data_fp32)

        return loss



from __future__ import annotations
from collections.abc import Iterable, Iterator
from datetime import datetime
import random
from typing import Any, TypeVar

import numpy as np
import torch
import torch.nn as nn

__all__ = ["count_parameters", "random_seed", "inf_generator", "Event"]


def count_parameters(model: nn.Module) -> int:
    """Count all parameters (that have a gradient) in the given model"""
    return sum(p.numel() for p in model.parameters() if p.requires_grad)


def random_seed(seed_value: int, *, use_cuda: bool) -> None:
    np.random.seed(seed_value)  # cpu vars
    torch.manual_seed(seed_value)  # cpu  vars
    random.seed(seed_value)  # Python
    if use_cuda:
        torch.cuda.manual_seed(seed_value)  # type: ignore
        torch.cuda.manual_seed_all(seed_value)  # type: ignore
        torch.backends.cudnn.deterministic = True  # type: ignore
        torch.backends.cudnn.benchmark = False  # type: ignore


T = TypeVar("T")


def inf_generator(iterable: Iterable[T]) -> Iterator[T]:
    """Get DataLoaders in a single infinite loop.

    for i, (x, y) in enumerate(inf_generator(train_loader))
    """
    iterator = iter(iterable)
    # try to take one element to ensure that the iterator is not empty
    first_value = next(iterator, None)
    if first_value is not None:
        yield first_value
    else:
        raise RuntimeError("The given iterable is empty.")
    while True:
        try:
            yield next(iterator)
        except StopIteration:
            iterator = iter(iterable)


class Event:
    """Emulates torch.cuda.Event, but supports running on a CPU too.

    :example:

    >>> from ranzen.torch import Event
    >>> with Event() as event:
    >>>     y = some_nn_module(x)
    >>> print(event.time)
    """

    def __init__(self):
        self.time = 0.0
        self._cuda = torch.cuda.is_available()  # type: ignore
        self._event_start: torch.cuda.Event | datetime  # type: ignore

    def __enter__(self) -> Event:
        """Mark a time.

        Mimics torch.cuda.Event.
        """
        if self._cuda:
            self._event_start = torch.cuda.Event(enable_timing=True)  # type: ignore
            self._event_start.record()  # type: ignore
        else:
            self._event_start = datetime.now()
        return self

    def __exit__(self, *args: Any) -> None:
        if self._cuda:
            event_end = torch.cuda.Event(enable_timing=True)  # type: ignore
            event_end.record()
            torch.cuda.synchronize()  # type: ignore
            assert isinstance(self._event_start, torch.cuda.Event)  # type: ignore
            self.time = self._event_start.elapsed_time(event_end)  # type: ignore
        else:
            assert isinstance(self._event_start, datetime)
            self.time = datetime.now().microsecond - self._event_start.microsecond

    def __repr__(self) -> str:
        return f"Event of duration: {self.time}"



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['ranzen',
 'ranzen.hydra',
 'ranzen.pl',
 'ranzen.torch',
 'ranzen.torch.optimizers',
 'ranzen.torch.transforms']

package_data = \
{'': ['*']}

extras_require = \
{'all': ['pytorch-lightning>=1.4.2,<2.0.0',
         'hydra-core>=1.1.0,<2.0.0',
         'neoconfigen>=2.0.0,<3.0.0',
         'pandas>=1.3.1,<2.0.0',
         'wandb>=0.12.0,<0.13.0'],
 'hydra': ['hydra-core>=1.1.0,<2.0.0', 'neoconfigen>=2.0.0,<3.0.0'],
 'pl': ['pytorch-lightning>=1.4.2,<2.0.0', 'tqdm>=4.62.0,<5.0.0'],
 'torch': ['torch>=1.8,<2.0', 'numpy>=1.20.3,<2.0.0'],
 'wandb': ['pandas>=1.3.1,<2.0.0', 'wandb>=0.12.0,<0.13.0']}

setup_kwargs = {
    'name': 'ranzen',
    'version': '1.1.2',
    'description': 'A toolkit facilitating machine-learning experimentation.',
    'long_description': '# ranzen\n\nA python toolkit facilitating machine-learning experimentation.\n\n[Documentation](https://wearepal.ai/ranzen/)\n\n## Install\n\nRun\n```\npip install ranzen\n```\n\nor install directly from GitHub:\n```\npip install git+https://github.com/predictive-analytics-lab/ranzen.git@main\n```\n',
    'author': 'PAL',
    'author_email': 'info@predictive-analytics-lab.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/predictive-analytics-lab/ranzen',
    'packages': packages,
    'package_data': package_data,
    'extras_require': extras_require,
    'python_requires': '>=3.7.1,<4.0.0',
}


setup(**setup_kwargs)
